﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using SlimDX;

namespace Models.Aqo
{
    public class AFPFile
    {
        public string                   FileName                { get; set; }

        // afp
        public string                   afp                     { get; set; }
        public byte[]                   Unknown0                { get; set; }
        public uint                     Size1                   { get; set; }
        public uint                     Size2                   { get; set; }
        public uint                     Size3                   { get; set; }

        // aqo
        public string                   aqo                     { get; set; }

        // VTBF
        public string                   VTBF                    { get; set; }
        public uint                     Unknown1                { get; set; }

        // AQGF
        public string                   AQGF                    { get; set; }
        public uint                     Unknown2                { get; set; }

        // chunks
        public List<vtc0>               VTC0                    { get; set; }
        public Dictionary<string, BaseData[]> Dictionary        { get; set; }
        public ROOT                     ROOT                    { get { return GetFirst<ROOT>("ROOT"); } }
        public OBJC                     OBJC                    { get { return GetFirst<OBJC>("OBJC"); } }
        public VSET                     VSET                    { get { return GetFirst<VSET>("VSET"); } }
        public VTXE[]                   VTXE                    { get { return GetArray<VTXE>("VTXE"); } }
        public VTXL[]                   VTXL                    { get { return GetArray<VTXL>("VTXL"); } }
        public PSET                     PSET                    { get { return GetFirst<PSET>("PSET"); } }
        public MESH                     MESH                    { get { return GetFirst<MESH>("MESH"); } }
        public MATE                     MATE                    { get { return GetFirst<MATE>("MATE"); } }
        public REND                     REND                    { get { return GetFirst<REND>("REND"); } }
        public SHAD                     SHAD                    { get { return GetFirst<SHAD>("SHAD"); } }
        public SHAP[]                   SHAP                    { get { return GetArray<SHAP>("SHAP"); } }
        public TSET                     TSET                    { get { return GetFirst<TSET>("TSET"); } }
        public TSTA                     TSTA                    { get { return GetFirst<TSTA>("TSTA"); } }
        public TEXF                     TEXF                    { get { return GetFirst<TEXF>("TEXF"); } }
        public UNRM                     UNRM                    { get { return GetFirst<UNRM>("UNRM"); } }
        public NDTR                     NDTR                    { get { return GetFirst<NDTR>("NDTR"); } }
        public NODE                     NODE                    { get { return GetFirst<NODE>("NODE"); } }
        public NODO                     NODO                    { get { return GetFirst<NODO>("NODO"); } }

        public AFPFile()
        {
            VTC0   = new List<vtc0>();
        }

        public Dictionary<string, BaseData[]> AsDictionary()
        {
            return VTC0
                .Select(i => i.DataObj)
                .GroupBy(i => i.GetType().Name)
                .ToDictionary(i => i.Key, i => i.ToArray());
        }

        public void Extract(string dir)
        {
            if(!Directory.Exists(dir))
                Directory.CreateDirectory(dir);

            int n   = 0;

            foreach(var i in VTC0)
            {
                var name= string.Format("{0:D3}_{1:X8}.PSO2_{2}", n++, i.Position, i.SubType);
                var file= Path.Combine(dir, name);

                File.WriteAllBytes(file, i.Data);
            }
        }

        public T GetFirst<T>(string tag) where T : BaseData
        {
            return Dictionary.ContainsKey(tag) ? (T)Dictionary[tag].First() : null;
        }

        public T[] GetArray<T>(string tag)
        {
            return Dictionary.ContainsKey(tag) ? (T[])Dictionary[tag].Cast<T>().ToArray() : null;
        }

        public static AFPFile FromFile(string file)
        {
            using(var r= new AFPReader(file))
                return r.Read();
        }

        public static AFPFile FromStream(Stream s)
        {
            using(var r= new AFPReader(s))
                return r.Read();
        }

        public static void ToFile(string file, AFPFile data)
        {
            using(var w= new AFPWriter(file))
                w.Write(data);
        }
   
        #region Verify
        public void Verify()
        {
            System.Diagnostics.Debug.Print(FileName);

            var OBJC= this.OBJC;
            var VSET= this.VSET;
            var VTXE= this.VTXE;
            var VTXL= this.VTXL;
            var PSET= this.PSET;
            var MESH= this.MESH;
            var MATE= this.MATE;
            var REND= this.REND;
            var SHAD= this.SHAD;
            var SHAP= this.SHAP;
            var TSET= this.TSET;
            var TSTA= this.TSTA;
            var TEXF= this.TEXF;
            var UNRM= this.UNRM;

          //Assert(MESH.Data.Length < 4);;

            if(null != PSET)    Assert(VSET.Data.Length == PSET.Data.Length);
                                Assert(VTXE.Length == VTXL.Length);
                                Assert(OBJC._0816 == VSET.Data.Length);
                                Assert(OBJC._0924 == VSET.Data.Length);
                                Assert(OBJC._0925 == VSET.Data.Length);
            if(null != PSET)    Assert(OBJC._0917 == PSET.Data.Length);
            if(null != SHAD)    Assert(OBJC._081A == SHAD.Data.Length);
                              //Assert(OBJC._081C == SHAP.Length);
            if(null != MATE)    Assert(OBJC._0818 == MATE.Data.Length);
            if(null != REND)    Assert(OBJC._0819 == REND.Data.Length);
                              //Assert(OBJC._081B == TSTA.Data.Length); // pl_dr_fc_001でエラー
                              //Assert(OBJC._081D == TEXF.Data.Length); // pl_dr_fc_001でエラー

            var totalindex  = PSET.Data.Sum(i => i.NumIndex1);
            var totalvertex = VSET.Data.Sum(i => i.NumVert);

            Assert(totalindex  == OBJC.TotalNumIndex);
            Assert(totalvertex == OBJC.TotalNumVert);

            foreach(var i in MESH.Data)
            {
                var vset    = VSET.Data[i.IndexVSET];
                var pset    = PSET.Data[i.IndexPSET];
                var vtxe    = VTXE[i.IndexVSET];
                var vtxl    = VTXL[i.IndexVSET];

                Assert(vset.NumVert * vset.VertSize == vtxl.VertexData.Length);
                Assert(pset.NumIndex1 == pset.Indices.Length);
                Assert(pset.NumIndex2 == pset.Indices.Length);
            }

            if(null != MESH)
            foreach(var i in MESH.Data)
            {
                Assert(i._08B1 < MATE.Data.Length); // IndexMATE
                Assert(i._08B2 < REND.Data.Length); // IndexREND
                Assert(i._08B3 < SHAD.Data.Length); // IndexSHAD
                Assert(i._08B4 < TSET.Data.Length); // IndexTSET
                Assert(i._08C0 < VSET.Data.Length); // IndexVSET
                Assert(i._08C1 < PSET.Data.Length); // IndexPSET

              //Assert(i._08B1 < 6);
              //Assert(i._08B5 < 86);
            }

            if(null != TSET)
            foreach(var i in TSET.Data)
            {
                Assert(i._0871 == i._8875.Count(j => j != 0xFFFFFFFF));
            }

            foreach(var i in VTC0)
            {
                foreach(var j in i.DataObj.Params)
                {
                    if(j.Value is uint)
                    {
                        //ObjExporter.AddValueContains(j.Tag, (uint)j.Value);

                        #if true
                        switch(j.ID)
                        {
                        case 0x0810: Verify     ((uint)j.Value, 3114);                                                           break; // [1]
                        case 0x0811: Verify     ((uint)j.Value, 164);                                                            break; // [1]
                        case 0x0912: Verify     ((uint)j.Value, 18, 17, 12, 20, 11, 10, 9, 22, 21, 19, 13, 27, 31, 23, 29, 26);  break; // [16]
                        case 0x0813: Verify     ((uint)j.Value, 76, 56, 32, 52, 44, 60, 68, 48, 84);                             break; // [9]
                        case 0x0914: VerifyRange((uint)j.Value, 3, 32000);                                                       break; // [1561]
                        case 0x0815: VerifyRange((uint)j.Value, 3, 32000);                                                       break; // [1237]
                        case 0x0816: VerifyRange((uint)j.Value, 1, 118);                                                         break; // [64]
                        case 0x0924: VerifyRange((uint)j.Value, 1, 118);                                                         break; // [64]
                        case 0x0925: VerifyRange((uint)j.Value, 1, 118);                                                         break; // [64]
                        case 0x0917: VerifyRange((uint)j.Value, 1, 118);                                                         break; // [64]
                        case 0x0818: Verify     ((uint)j.Value, 1, 2, 3, 4, 5, 6, 7);                                            break; // [7]
                        case 0x0819: Verify     ((uint)j.Value, 2, 1, 3, 4, 6, 5);                                               break; // [6]
                        case 0x081A: Verify     ((uint)j.Value, 1, 2, 3);                                                        break; // [3]
                        case 0x081B: Verify     ((uint)j.Value, 3, 4, 5, 7, 2, 6, 0, 8, 9);                                      break; // [9]
                        case 0x081C: Verify     ((uint)j.Value, 1, 2, 3, 4);                                                     break; // [4]
                        case 0x081D: Verify     ((uint)j.Value, 3, 4, 5, 7, 2, 6, 0, 8, 9);                                      break; // [9]
                        case 0x09B6: Verify     ((uint)j.Value, 76, 56, 32, 52, 24, 48, 44, 64, 40, 60, 68, 84);                 break; // [12]
                        case 0x09BF: Verify     ((uint)j.Value, 7, 5, 3, 6, 8);                                                  break; // [5]
                        case 0x09B9: VerifyRange((uint)j.Value, 3, 8454);                                                        break; // [1393]
                        case 0x09C4: Verify     ((uint)j.Value, 0)       ;                                                       break; // [1]
                        case 0x08BD: Verify     ((uint)j.Value, 1, 2, 3, 0, 6, 9, 13, 7, 5, 8, 4, 10, 12, 16, 11, 14, 15);       break; // [17]
                        case 0x09C8: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09CC: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09C9: VerifyRange((uint)j.Value, 0, 1414);                                                        break; // [63]
                        case 0x09D0: Verify     ((uint)j.Value, 0, 1, 2, 33, 32, 11, 16, 3, 4, 17);                              break; // [10]
                        case 0x09D1: Verify     ((uint)j.Value, 3, 4, 7, 2, 5);                                                  break; // [5]
                        case 0x09D2: Verify     ((uint)j.Value, 0, 12, 28, 40, 52, 64, 68, 24, 36, 48, 44, 16, 32, 56, 60, 76);  break; // [16]
                        case 0x09D3: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09C6: Verify     ((uint)j.Value, 8448);                                                           break; // [1]
                      //case 0x09BB: Verify     ((uint)j.Value, 1);                                                              break; // [1]
                        case 0x09BB: Verify     ((uint)j.Value, 0, 1);                                                           break; // [1]
                        case 0x09BC: VerifyRange((uint)j.Value, 3, 17327);                                                       break; // [2111]
                        case 0x09B7: VerifyRange((uint)j.Value, 3, 17327);                                                       break; // [2111]
                        case 0x09C5: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09B0: Verify     ((uint)j.Value, 1048594, 17, 1048588, 1048596, 1048585, 1048586, 9, 20, 10, 1048593, 12, 18, 42, 50);                                                       break; // [14]
                        case 0x09C7: VerifyRange((uint)j.Value, 23680, 27520);                                                   break; // [196]
                        case 0x08B1: Verify     ((uint)j.Value, 0, 1, 2, 3, 4, 5, 6);                                            break; // [7]   AlphaBlend?
                        case 0x08B2: Verify     ((uint)j.Value, 0, 1, 2, 3, 4, 5);                                               break; // [6]   AlphaBlend?
                        case 0x08B3: Verify     ((uint)j.Value, 0, 1, 2);                                                        break; // [3]
                        case 0x08B4: Verify     ((uint)j.Value, 0, 1, 2, 3);                                                     break; // [4]
                        case 0x08B5: VerifyRange((uint)j.Value, 0, 86);                                                          break; // [49]
                        case 0x08C0: VerifyRange((uint)j.Value, 0, 117);                                                         break; // [118]
                        case 0x08C1: VerifyRange((uint)j.Value, 0, 117);                                                         break; // [118]
                        case 0x08CD: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09C2: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0934: Verify     ((uint)j.Value, 0, 2, 4);                                                        break; // [3]
                        case 0x0937: Verify     ((uint)j.Value, 90, 100, 105, 99, 102, 101, 106, 104, 97, 96, 95, 98, 110);      break; // [13]
                        case 0x0938: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0940: Verify     ((uint)j.Value, 511);                                                            break; // [1]
                        case 0x0941: Verify     ((uint)j.Value, 3, 1);                                                           break; // [2]
                        case 0x0942: Verify     ((uint)j.Value, 2, 1);                                                           break; // [2]
                        case 0x0943: Verify     ((uint)j.Value, 1, 0);                                                           break; // [2]
                        case 0x0944: Verify     ((uint)j.Value, 5);                                                              break; // [1]
                        case 0x0945: Verify     ((uint)j.Value, 6, 2);                                                           break; // [2]
                        case 0x0946: Verify     ((uint)j.Value, 1);                                                              break; // [1]
                        case 0x0947: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0948: Verify     ((uint)j.Value, 5);                                                              break; // [1]
                        case 0x0949: Verify     ((uint)j.Value, 6);                                                              break; // [1]
                        case 0x094A: Verify     ((uint)j.Value, 1);                                                              break; // [1]
                        case 0x094B: Verify     ((uint)j.Value, 1, 0);                                                           break; // [2]
                        case 0x094C: Verify     ((uint)j.Value, 5);                                                              break; // [1]
                        case 0x094D: Verify     ((uint)j.Value, 0, 128);                                                         break; // [2]
                        case 0x094E: Verify     ((uint)j.Value, 1);                                                              break; // [1]
                        case 0x094F: Verify     ((uint)j.Value, 4);                                                              break; // [1]
                        case 0x0950: Verify     ((uint)j.Value, 0, 1);                                                           break; // [2]
                        case 0x0990: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0993: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0960: Verify     ((uint)j.Value, 22);                                                             break; // [1]
                        case 0x0961: Verify     ((uint)j.Value, 0, 1, 2, 3, 6);                                                  break; // [5]
                        case 0x0962: Verify     ((uint)j.Value, 0, 4294967295, 1);                                               break; // [3]
                        case 0x0964: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0965: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0966: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0967: Verify     ((uint)j.Value, 1, 3);                                                           break; // [2]
                        case 0x0968: Verify     ((uint)j.Value, 1, 3);                                                           break; // [2]
                        case 0x0969: Verify     ((uint)j.Value, 1);                                                              break; // [1]
                        case 0x0970: Verify     ((uint)j.Value, 0, 1);                                                           break; // [2]
                        case 0x0871: Verify     ((uint)j.Value, 3, 4, 1, 2, 0);                                                  break; // [5]
                        case 0x0972: Verify     ((uint)j.Value, 0, 2);                                                           break; // [2]
                        case 0x0973: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x0974: Verify     ((uint)j.Value, 0);                                                              break; // [1]
                        case 0x09DA: VerifyRange((uint)j.Value, 8, 1624);                                                        break; // [287]
                        case 0x09DC: VerifyRange((uint)j.Value, 16, 3760);                                                       break; // [319]
                        }
                        #endif
                    }
                }
            }
        }

        public static void VerifyRange(uint value, uint min, uint max)
        {
            if(value < min)
                throw new Exception();

            if(value > max)
                throw new Exception();
        }

        public static void Verify(uint value, params uint[] values)
        {
            if(!values.Contains(value))
                throw new Exception();
        }
        #endregion

        #region Dump
        public void Dump(string file)
        {
            using(var w= new StreamWriter(file, false, Encoding.UTF8))
            {
                Dump(w);
                w.Flush();
                w.BaseStream.SetLength(w.BaseStream.Position);
            }
        }

        public void Dump(StreamWriter sw)
        {
            var dic1= VTC0.Select(i => i.DataObj).GroupBy(i => i.GetType().Name).ToDictionary(i => i.Key, i => i.ToArray());

            foreach(var i in dic1)
            {
                sw.WriteLine("{0}", i.Key);

                int idx = 0;

                foreach(var j in i.Value)
                {
                    sw.WriteLine("  {0}[{1}]", i.Key, idx++);

                    Dump(sw, j);
                }
            }
        }

        public void Dump(StreamWriter sw, BaseData bd)
        {
            var type= bd.GetType();
            var sb  = new StringBuilder();

            foreach(var i in type.GetProperties().Where(i => i.Name.StartsWith("_")))
            {
                sb.Length= 0;
                var name= i.Name.Trim('_');
                var val = i.GetValue(bd, null);

                sb.AppendFormat("    {0:X4}:{1}\t", name, AFPUtils.GetParamName(Convert.ToUInt16(name, 16)).PadRight(16));

                     if(val == null)        {}
                else if(val is string)      sb.Append('"').Append(val).Append('"');
                else if(val is int)         sb.AppendFormat("{0:X8}/{0:N0}", val);
                else if(val is uint)        sb.AppendFormat("{0:X8}/{0:N0}", val);
                else if(val is ushort)      sb.AppendFormat("{0:X4}/{0:N0}", val);
                else if(val is float)       sb.AppendFormat("{0:F0}", val);
                else if(val is ushort[])    sb.Append(ToString((ushort[])val, 8, x => x.ToString()));
                else if(val is float[])     sb.Append(ToString((float[]) val, 8, x => x.ToString()));
                else if(val is int[])       sb.Append(ToString((int[])   val, 8, x => x.ToString()));
                else if(val is uint[])      sb.Append(ToString((uint[])  val, 8, x => x.ToString()));
                else if(val is byte[])      sb.Append(ToString((byte[])  val, 8, x => x.ToString("X2")));
                else if(val is Vector3)     sb.Append(val.ToString());
                else if(val is Vector4)     sb.Append(val.ToString());
                else { throw new Exception(); }

                sw.WriteLine(sb.ToString());
            }

            var df  = type.GetProperty("Data");

            if(df == null)
                return;

            var data= (Array)df.GetValue(bd, null);
            var idx = 0;

            sw.WriteLine("    NumData: {0}", data.Length);

            foreach(var i in data)
            {
                sw.WriteLine("    {0}[{1}]", i.GetType().Name, idx++);

                Dump(sw, (ParamSet)i);
            }
        }

        public void Dump(StreamWriter sw, ParamSet ps)
        {
            var type= ps.GetType();
            var sb  = new StringBuilder();

            foreach(var i in type.GetProperties().Where(i => i.Name.StartsWith("_")))
            {
                sb.Length= 0;
                var name= i.Name.Trim('_');
                var val = i.GetValue(ps, null);

                sb.AppendFormat("      {0:X4}:{1}\t", name, AFPUtils.GetParamName(Convert.ToUInt16(name, 16)).PadRight(16));

                     if(val == null)        {}
                else if(val is string)      sb.Append('"').Append(val).Append('"');
                else if(val is int)         sb.AppendFormat("{0:X8}/{0:N0}", val);
                else if(val is uint)        sb.AppendFormat("{0:X8}/{0:N0}", val);
                else if(val is ushort)      sb.AppendFormat("{0:X4}/{0:N0}", val);
                else if(val is float)       sb.AppendFormat("{0:F0}", val);
                else if(val is ushort[])    sb.Append(ToString((ushort[])val, 8, x => x.ToString()));
                else if(val is float[])     sb.Append(ToString((float[]) val, 8, x => x.ToString()));
                else if(val is int[])       sb.Append(ToString((int[])   val, 8, x => x.ToString()));
                else if(val is uint[])      sb.Append(ToString((uint[])  val, 8, x => x.ToString()));
                else if(val is byte[])      sb.Append(ToString((byte[])  val, 8, x => x.ToString("X2")));
                else if(val is Vector3)     sb.Append(val.ToString());
                else if(val is Vector4)     sb.Append(val.ToString());
                else if(val is Matrix)      sb.Append(val.ToString());
                else { throw new Exception(); }

                sw.WriteLine(sb.ToString());
            }
        }
        #endregion

        public static string ToString<T>(T[] a, int count, Func<T, string> func)
        {
            #if false
            return a.Length > count
                ? string.Format("[{0}]\t{{ {1}, ... }}", a.Length, string.Join(", ", a.Take(count).Select(i => func(i))))
                : string.Format("[{0}]\t{{ {1} }}",      a.Length, string.Join(", ", a.Select(i => func(i))));
            #else
            return string.Format("[{0}]\t{{ {1} }}",      a.Length, string.Join(", ", a.Select(i => func(i))));
            #endif
        }

        public static void Assert(bool b)
        {
            if(!b)
                throw new Exception();
        }
    }

    public class vtc0
    {
        public long                     Position                { get; set; }
        public string                   Type                    { get; set; }
        public int                      Size                    { get; set; }
        public string                   SubType                 { get; set; }
        public byte[]                   Data                    { get; set; }
        public BaseData                 DataObj                 { get; set; }

        public vtc0()
        {
            Type     = "vtc0";
        }

        public vtc0(BaseData data)
        {
            Type     = "vtc0";
            SubType  = data.GetType().Name;
            DataObj = data;
        }
    }

    public class ParamSet
    {
        public List<Param>              Params                  { get; set; }

        public ParamSet()
        {
            Params  = new List<Param>();
        }

        public T GetParamAs<T>(ushort id)
        {
            var p   = Params.FirstOrDefault(i => i.ID == id);

            return p == null ? default(T) : (T)p.Value;
        }

        public void SetParam<T>(ushort id, T value)
        {
            Params.First(i => i.ID == id).Value = value;
        }
    }

    public class BaseData : ParamSet
    {
        public ushort                   Head                    { get; set; }

        public void InitParams(byte[] data)
        {
            Head    = BitConverter.ToUInt16(data, 0);
            Params  = Param.ParseParams(data).ToList();
        }
    }

    public class ROOT : BaseData
    {
        public string                   _0200                   { get { return GetParamAs<string>  (0x0200); } set { SetParam(0x0200, value); } }

        public ROOT(byte[] data)
        {
            InitParams(data);
        }
    }

    public class OBJC : BaseData
    {
        public int                      _0810                   { get { return GetParamAs<int>     (0x0810); } set { SetParam(0x0810, value); } }
        public int                      _0811                   { get { return GetParamAs<int>     (0x0811); } set { SetParam(0x0811, value); } }
        public uint                     _0912                   { get { return GetParamAs<uint>    (0x0912); } set { SetParam(0x0912, value); } }
        public int                      _0813                   { get { return GetParamAs<int>     (0x0813); } set { SetParam(0x0813, value); } }
        public uint                     _0914                   { get { return GetParamAs<uint>    (0x0914); } set { SetParam(0x0914, value); } }
        public int                      _0815                   { get { return GetParamAs<int>     (0x0815); } set { SetParam(0x0815, value); } }
        public int                      _0816                   { get { return GetParamAs<int>     (0x0816); } set { SetParam(0x0816, value); } }
        public uint                     _0924                   { get { return GetParamAs<uint>    (0x0924); } set { SetParam(0x0924, value); } }
        public uint                     _0925                   { get { return GetParamAs<uint>    (0x0925); } set { SetParam(0x0925, value); } }
        public uint                     _0917                   { get { return GetParamAs<uint>    (0x0917); } set { SetParam(0x0917, value); } }
        public int                      _0818                   { get { return GetParamAs<int>     (0x0818); } set { SetParam(0x0818, value); } }
        public int                      _0819                   { get { return GetParamAs<int>     (0x0819); } set { SetParam(0x0819, value); } }
        public int                      _081A                   { get { return GetParamAs<int>     (0x081A); } set { SetParam(0x081A, value); } }
        public int                      _081B                   { get { return GetParamAs<int>     (0x081B); } set { SetParam(0x081B, value); } }
        public int                      _081C                   { get { return GetParamAs<int>     (0x081C); } set { SetParam(0x081C, value); } }
        public int                      _081D                   { get { return GetParamAs<int>     (0x081D); } set { SetParam(0x081D, value); } }
        public Vector3                  _4A1E                   { get { return GetParamAs<Vector3> (0x4A1E); } set { SetParam(0x4A1E, value); } }
        public float                    _0A1F                   { get { return GetParamAs<float>   (0x0A1F); } set { SetParam(0x0A1F, value); } }
        public Vector3                  _4A20                   { get { return GetParamAs<Vector3> (0x4A20); } set { SetParam(0x4A20, value); } }
        public Vector3                  _4A21                   { get { return GetParamAs<Vector3> (0x4A21); } set { SetParam(0x4A21, value); } }

        public int                      Fixed1_3114             { get { return GetParamAs<int>     (0x0810); } set { SetParam(0x0810, value); } }
        public int                      Fixed2_164              { get { return GetParamAs<int>     (0x0811); } set { SetParam(0x0811, value); } }
      //public uint                     _0912                   { get { return GetParamAs<uint>    (0x0912); } set { SetParam(0x0912, value); } }
        public int                      VertSize                { get { return GetParamAs<int>     (0x0813); } set { SetParam(0x0813, value); } }
        public uint                     TotalNumIndex           { get { return GetParamAs<uint>    (0x0914); } set { SetParam(0x0914, value); } }
        public int                      TotalNumVert            { get { return GetParamAs<int>     (0x0815); } set { SetParam(0x0815, value); } }
        public int                      NumVSET1                { get { return GetParamAs<int>     (0x0816); } set { SetParam(0x0816, value); } }
        public uint                     NumVSET2                { get { return GetParamAs<uint>    (0x0924); } set { SetParam(0x0924, value); } }
        public uint                     NumVSET3                { get { return GetParamAs<uint>    (0x0925); } set { SetParam(0x0925, value); } }
        public uint                     NumVSET4                { get { return GetParamAs<uint>    (0x0917); } set { SetParam(0x0917, value); } }
        public int                      NnuMATE                 { get { return GetParamAs<int>     (0x0818); } set { SetParam(0x0818, value); } }
        public int                      NumREND                 { get { return GetParamAs<int>     (0x0819); } set { SetParam(0x0819, value); } }
        public int                      NumSHAD                 { get { return GetParamAs<int>     (0x081A); } set { SetParam(0x081A, value); } }
        public int                      NumTSTA                 { get { return GetParamAs<int>     (0x081B); } set { SetParam(0x081B, value); } }
        public int                      NumSHAP                 { get { return GetParamAs<int>     (0x081C); } set { SetParam(0x081C, value); } }
        public int                      NumTEXF                 { get { return GetParamAs<int>     (0x081D); } set { SetParam(0x081D, value); } }
        public Vector3                  Center                  { get { return GetParamAs<Vector3> (0x4A1E); } set { SetParam(0x4A1E, value); } }
      //public float                    _0A1F                   { get { return GetParamAs<float>   (0x0A1F); } set { SetParam(0x0A1F, value); } }
        public Vector3                  BBCenter                { get { return GetParamAs<Vector3> (0x4A20); } set { SetParam(0x4A20, value); } }
        public Vector3                  BBExtents               { get { return GetParamAs<Vector3> (0x4A21); } set { SetParam(0x4A21, value); } }

        public OBJC(byte[] data)
        {
            InitParams(data);
        }
    }

    public class VSET : BaseData
    {
        public VSETData[]               Data                    { get; set; }

        public VSET(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public VSET(IEnumerable<VSETData> data)
        {
            Params.Clear();

            var first   = true;

            foreach(var i in data)
            {
                Params.Add(new Param((ushort)(first ? 0x00FC : 0x00FE)));
                Params.AddRange(i.Params);

                first   = false;
            }

            Params.Add(new Param(0x00FD));
        }

        public IEnumerable<VSETData> GenerateData
        {
            get
            {
                var data= new VSETData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x09B6:
                    case 0x09BF:
                    case 0x09B9:
                    case 0x09C4:
                    case 0x08BD:
                    case 0x06BE:
                    case 0x86BE:
                    case 0x09C8:
                    case 0x09CC:
                    case 0x09C9:
                    case 0x06CA:
                    case 0x86CA: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new VSETData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class VSETData : ParamSet
        {
            public uint                 _09B6                   { get { return GetParamAs<uint>    (0x09B6); } set { SetParam(0x09B6, value); } }
            public uint                 _09BF                   { get { return GetParamAs<uint>    (0x09BF); } set { SetParam(0x09BF, value); } }
            public uint                 _09B9                   { get { return GetParamAs<uint>    (0x09B9); } set { SetParam(0x09B9, value); } }
            public uint                 _09C4                   { get { return GetParamAs<uint>    (0x09C4); } set { SetParam(0x09C4, value); } }
            public int                  _08BD                   { get { return GetParamAs<int>     (0x08BD); } set { SetParam(0x08BD, value); } }
            public uint                 _09C8                   { get { return GetParamAs<uint>    (0x09C8); } set { SetParam(0x09C8, value); } }
            public uint                 _09CC                   { get { return GetParamAs<uint>    (0x09CC); } set { SetParam(0x09CC, value); } }
            public uint                 _09C9                   { get { return GetParamAs<uint>    (0x09C9); } set { SetParam(0x09C9, value); } }
            public ushort               _06BE                   { get { return GetParamAs<ushort>  (0x06BE); } set { SetParam(0x06BE, value); } }
            public ushort[]             _86BE                   { get { return GetParamAs<ushort[]>(0x86BE); } set { SetParam(0x86BE, value); } }
            public ushort               _06CA                   { get { return GetParamAs<ushort>  (0x06CA); } set { SetParam(0x06CA, value); } }
            public ushort[]             _86CA                   { get { return GetParamAs<ushort[]>(0x86CA); } set { SetParam(0x86CA, value); } }

            public uint                 VertSize                { get { return GetParamAs<uint>    (0x09B6); } set { SetParam(0x09B6, value); } }
            public uint                 NumElem                 { get { return GetParamAs<uint>    (0x09BF); } set { SetParam(0x09BF, value); } }
            public uint                 NumVert                 { get { return GetParamAs<uint>    (0x09B9); } set { SetParam(0x09B9, value); } }
            public uint                 Fixed1_0                { get { return GetParamAs<uint>    (0x09C4); } set { SetParam(0x09C4, value); } }
            public int                  NumBone                 { get { return GetParamAs<int>     (0x08BD); } set { SetParam(0x08BD, value); } }
            public uint                 Fixed2_0                { get { return GetParamAs<uint>    (0x09C8); } set { SetParam(0x09C8, value); } }
            public uint                 Fixed3_0                { get { return GetParamAs<uint>    (0x09CC); } set { SetParam(0x09CC, value); } }
          //public uint                 _09C9                   { get { return GetParamAs<uint>    (0x09C9); } set { SetParam(0x09C9, value); } }
          //public ushort               _06BE                   { get { return GetParamAs<ushort>  (0x06BE); } set { SetParam(0x06BE, value); } }
            public ushort[]             Bones                   { get { return GetParamAs<ushort[]>(0x86BE); } set { SetParam(0x86BE, value); } }
          //public ushort               _06CA                   { get { return GetParamAs<ushort>  (0x06CA); } set { SetParam(0x06CA, value); } }
          //public ushort[]             _86CA                   { get { return GetParamAs<ushort[]>(0x86CA); } set { SetParam(0x86CA, value); } }
        }
    }

    public class VTXE : BaseData
    {
        public VTXEData[]               Data                    { get; set; }

        public VTXE()
        {
        }

        public VTXE(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public VTXE(IEnumerable<VTXEData> data)
        {
            Params.Clear();

            var first   = true;

            foreach(var i in data)
            {
                Params.Add(new Param((ushort)(first ? 0x00FC : 0x00FE)));
                Params.AddRange(i.Params);

                first   = false;
            }

            Params.Add(new Param(0x00FD));
        }

        public IEnumerable<VTXEData> GenerateData
        {
            get
            {
                var data= new VTXEData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x09D0:
                    case 0x09D1:
                    case 0x09D2:
                    case 0x09D3: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new VTXEData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class VTXEData : ParamSet
        {
            public uint                 _09D0                   { get { return GetParamAs<uint>    (0x09D0); } set { SetParam(0x09D0, value); } }
            public uint                 _09D1                   { get { return GetParamAs<uint>    (0x09D1); } set { SetParam(0x09D1, value); } }
            public uint                 _09D2                   { get { return GetParamAs<uint>    (0x09D2); } set { SetParam(0x09D2, value); } }
            public uint                 _09D3                   { get { return GetParamAs<uint>    (0x09D3); } set { SetParam(0x09D3, value); } }

            public uint                 Usage                   { get { return GetParamAs<uint>    (0x09D0); } set { SetParam(0x09D0, value); } }
            public uint                 Type                    { get { return GetParamAs<uint>    (0x09D1); } set { SetParam(0x09D1, value); } }
            public uint                 Offset                  { get { return GetParamAs<uint>    (0x09D2); } set { SetParam(0x09D2, value); } }
            public uint                 Index                   { get { return GetParamAs<uint>    (0x09D3); } set { SetParam(0x09D3, value); } }
        }
    }

    public class VTXL : BaseData
    {
        public byte[]                   _89BA                   { get { return GetParamAs<byte[]>  (0x89BA); } set { SetParam(0x89BA, value); } }
        public byte[]                   VertexData              { get { return GetParamAs<byte[]>  (0x89BA); } set { SetParam(0x89BA, value); } }

        public VTXL()
        {
        }

        public VTXL(byte[] data)
        {
            InitParams(data);
        }
    }

    public class PSET : BaseData
    {
        public PSETData[]               Data                    { get; set; }

        public PSET(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public PSET(IEnumerable<PSET.PSETData> data)
        {
            Params.Clear();

            var first   = true;

            foreach(var i in data)
            {
                Params.Add(new Param((ushort)(first ? 0x00FC : 0x00FE)));
                Params.AddRange(i.Params);

                first   = false;
            }

            Params.Add(new Param(0x00FD));
        }

        public IEnumerable<PSETData> GenerateData
        {
            get
            {
                var data= new PSETData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x09C6:
                    case 0x09BB:
                    case 0x09BC:
                    case 0x09B7:
                    case 0x86B8:
                    case 0x09C5: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new PSETData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class PSETData : ParamSet
        {
            public uint                 _09C6                   { get { return GetParamAs<uint>    (0x09C6); } set { SetParam(0x09C6, value); } }
            public uint                 _09BB                   { get { return GetParamAs<uint>    (0x09BB); } set { SetParam(0x09BB, value); } }
            public uint                 _09BC                   { get { return GetParamAs<uint>    (0x09BC); } set { SetParam(0x09BC, value); } }
            public uint                 _09B7                   { get { return GetParamAs<uint>    (0x09B7); } set { SetParam(0x09B7, value); } }
            public ushort[]             _86B8                   { get { return GetParamAs<ushort[]>(0x86B8); } set { SetParam(0x86B8, value); } }
            public uint                 _09C5                   { get { return GetParamAs<uint>    (0x09C5); } set { SetParam(0x09C5, value); } }

            public uint                 Fixed1_8448             { get { return GetParamAs<uint>    (0x09C6); } set { SetParam(0x09C6, value); } }
            public uint                 Fixed2_1                { get { return GetParamAs<uint>    (0x09BB); } set { SetParam(0x09BB, value); } }
            public uint                 NumIndex1               { get { return GetParamAs<uint>    (0x09BC); } set { SetParam(0x09BC, value); } }
            public uint                 NumIndex2               { get { return GetParamAs<uint>    (0x09B7); } set { SetParam(0x09B7, value); } }
            public ushort[]             Indices                 { get { return GetParamAs<ushort[]>(0x86B8); } set { SetParam(0x86B8, value); } }
            public uint                 Fixed3_0                { get { return GetParamAs<uint>    (0x09C5); } set { SetParam(0x09C5, value); } }
        }
    }

    public class MESH : BaseData
    {
        public MESHData[]               Data                    { get; set; }

        public MESH(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public MESH(IEnumerable<MESH.MESHData> data)
        {
            Params.Clear();

            var first   = true;

            foreach(var i in data)
            {
                Params.Add(new Param((ushort)(first ? 0x00FC : 0x00FE)));
                Params.AddRange(i.Params);

                first   = false;
            }

            //Params.Add(new Param(0x00FD));
        }

        public IEnumerable<MESHData> GenerateData
        {
            get
            {
                var data= new MESHData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x09B0:
                    case 0x09C7:
                    case 0x08B1:
                    case 0x08B2:
                    case 0x08B3:
                    case 0x08B4:
                    case 0x08B5:
                    case 0x08C0:
                    case 0x08C1:
                    case 0x08CD:
                    case 0x09C2: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new MESHData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;;
            }
        }

        public class MESHData : ParamSet
        {
            public uint                 _09B0                   { get { return GetParamAs<uint>    (0x09B0); } set { SetParam(0x09B0, value); } }
            public uint                 _09C7                   { get { return GetParamAs<uint>    (0x09C7); } set { SetParam(0x09C7, value); } }
            public int                  _08B1                   { get { return GetParamAs<int>     (0x08B1); } set { SetParam(0x08B1, value); } }
            public int                  _08B2                   { get { return GetParamAs<int>     (0x08B2); } set { SetParam(0x08B2, value); } }
            public int                  _08B3                   { get { return GetParamAs<int>     (0x08B3); } set { SetParam(0x08B3, value); } }
            public int                  _08B4                   { get { return GetParamAs<int>     (0x08B4); } set { SetParam(0x08B4, value); } }
            public int                  _08B5                   { get { return GetParamAs<int>     (0x08B5); } set { SetParam(0x08B5, value); } }
            public int                  _08C0                   { get { return GetParamAs<int>     (0x08C0); } set { SetParam(0x08C0, value); } }
            public int                  _08C1                   { get { return GetParamAs<int>     (0x08C1); } set { SetParam(0x08C1, value); } }
            public int                  _08CD                   { get { return GetParamAs<int>     (0x08CD); } set { SetParam(0x08CD, value); } }
            public uint                 _09C2                   { get { return GetParamAs<uint>    (0x09C2); } set { SetParam(0x09C2, value); } }

          //public uint                 _09B0                   { get { return GetParamAs<uint>    (0x09B0); } set { SetParam(0x09B0, value); } }
          //public uint                 _09C7                   { get { return GetParamAs<uint>    (0x09C7); } set { SetParam(0x09C7, value); } }
            public int                  IndexMATE               { get { return GetParamAs<int>     (0x08B1); } set { SetParam(0x08B1, value); } }
            public int                  IndexREND               { get { return GetParamAs<int>     (0x08B2); } set { SetParam(0x08B2, value); } }
            public int                  IndexSHAD               { get { return GetParamAs<int>     (0x08B3); } set { SetParam(0x08B3, value); } }
            public int                  IndexTSET               { get { return GetParamAs<int>     (0x08B4); } set { SetParam(0x08B4, value); } }
          //public int                  _08B5                   { get { return GetParamAs<uint>    (0x08B5); } set { SetParam(0x08B5, value); } }
            public int                  IndexVSET               { get { return GetParamAs<int>     (0x08C0); } set { SetParam(0x08C0, value); } }
            public int                  IndexPSET               { get { return GetParamAs<int>     (0x08C1); } set { SetParam(0x08C1, value); } }
            public int                  Fixed1_0                { get { return GetParamAs<int>     (0x08CD); } set { SetParam(0x08CD, value); } }
            public uint                 Fixed2_0                { get { return GetParamAs<uint>    (0x09C2); } set { SetParam(0x09C2, value); } }
        }
    }

    public class MATE : BaseData
    {
        public MATEData[]               Data                    { get; set; }

        public MATE(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<MATEData> GenerateData
        {
            get
            {
                var data= new MATEData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x4A30:
                    case 0x4A31:
                    case 0x4A32:
                    case 0x4A33:
                    case 0x0934:
                    case 0x0A35:
                    case 0x0A36:
                    case 0x0937:
                    case 0x0938:
                    case 0x023A:
                    case 0x0239: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new MATEData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class MATEData : ParamSet
        {
            public Vector4              _4A30                   { get { return GetParamAs<Vector4> (0x4A30); } set { SetParam(0x4A30, value); } }
            public Vector4              _4A31                   { get { return GetParamAs<Vector4> (0x4A31); } set { SetParam(0x4A31, value); } }
            public Vector4              _4A32                   { get { return GetParamAs<Vector4> (0x4A32); } set { SetParam(0x4A32, value); } }
            public Vector4              _4A33                   { get { return GetParamAs<Vector4> (0x4A33); } set { SetParam(0x4A33, value); } }
            public uint                 _0934                   { get { return GetParamAs<uint>    (0x0934); } set { SetParam(0x0934, value); } }
            public float                _0A35                   { get { return GetParamAs<float>   (0x0A35); } set { SetParam(0x0A35, value); } }
            public float                _0A36                   { get { return GetParamAs<float>   (0x0A36); } set { SetParam(0x0A36, value); } }
            public uint                 _0937                   { get { return GetParamAs<uint>    (0x0937); } set { SetParam(0x0937, value); } }
            public uint                 _0938                   { get { return GetParamAs<uint>    (0x0938); } set { SetParam(0x0938, value); } }
            public string               _023A                   { get { return GetParamAs<string>  (0x023A); } set { SetParam(0x023A, value); } }
            public string               _0239                   { get { return GetParamAs<string>  (0x0239); } set { SetParam(0x0239, value); } }
        }
    }

    public class REND : BaseData
    {
        public RENDData[]               Data                    { get; set; }

        public REND(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<RENDData> GenerateData
        {
            get
            {
                var data= new RENDData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0940:
                    case 0x0941:
                    case 0x0942:
                    case 0x0943:
                    case 0x0944:
                    case 0x0945:
                    case 0x0946:
                    case 0x0947:
                    case 0x0948:
                    case 0x0949:
                    case 0x094A:
                    case 0x094B:
                    case 0x094C:
                    case 0x094D:
                    case 0x094E:
                    case 0x094F:
                    case 0x0950: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new RENDData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class RENDData : ParamSet
        {
            public uint                 _0940                   { get { return GetParamAs<uint>    (0x0940); } set { SetParam(0x0940, value); } }
            public uint                 _0941                   { get { return GetParamAs<uint>    (0x0941); } set { SetParam(0x0941, value); } }
            public uint                 _0942                   { get { return GetParamAs<uint>    (0x0942); } set { SetParam(0x0942, value); } }
            public uint                 _0943                   { get { return GetParamAs<uint>    (0x0943); } set { SetParam(0x0943, value); } }
            public uint                 _0944                   { get { return GetParamAs<uint>    (0x0944); } set { SetParam(0x0944, value); } }
            public uint                 _0945                   { get { return GetParamAs<uint>    (0x0945); } set { SetParam(0x0945, value); } }
            public uint                 _0946                   { get { return GetParamAs<uint>    (0x0946); } set { SetParam(0x0946, value); } }
            public uint                 _0947                   { get { return GetParamAs<uint>    (0x0947); } set { SetParam(0x0947, value); } }
            public uint                 _0948                   { get { return GetParamAs<uint>    (0x0948); } set { SetParam(0x0948, value); } }
            public uint                 _0949                   { get { return GetParamAs<uint>    (0x0949); } set { SetParam(0x0949, value); } }
            public uint                 _094A                   { get { return GetParamAs<uint>    (0x094A); } set { SetParam(0x094A, value); } }
            public uint                 _094B                   { get { return GetParamAs<uint>    (0x094B); } set { SetParam(0x094B, value); } }
            public uint                 _094C                   { get { return GetParamAs<uint>    (0x094C); } set { SetParam(0x094C, value); } }
            public uint                 _094D                   { get { return GetParamAs<uint>    (0x094D); } set { SetParam(0x094D, value); } }
            public uint                 _094E                   { get { return GetParamAs<uint>    (0x094E); } set { SetParam(0x094E, value); } }
            public uint                 _094F                   { get { return GetParamAs<uint>    (0x094F); } set { SetParam(0x094F, value); } }
            public uint                 _0950                   { get { return GetParamAs<uint>    (0x0950); } set { SetParam(0x0950, value); } }

            public uint                 AlpheBlendEnable        { get { return GetParamAs<uint>    (0x0943); } set { SetParam(0x0943, value); } }
            public uint                 AlpaRef                 { get { return GetParamAs<uint>    (0x094D); } set { SetParam(0x094D, value); } }
        }
    }

    public class SHAD : BaseData
    {
        public SHADData[]               Data                    { get; set; }

        public SHAD(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<SHADData> GenerateData
        {
            get
            {
                var data= new SHADData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0990:
                    case 0x0291:
                    case 0x0292:
                    case 0x0993: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new SHADData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class SHADData : ParamSet
        {
            public uint                 _0990                   { get { return GetParamAs<uint>    (0x0990); } set { SetParam(0x0990, value); } }
            public string               _0291                   { get { return GetParamAs<string>  (0x0291); } set { SetParam(0x0291, value); } }
            public string               _0292                   { get { return GetParamAs<string>  (0x0292); } set { SetParam(0x0292, value); } }
            public uint                 _0993                   { get { return GetParamAs<uint>    (0x0993); } set { SetParam(0x0993, value); } }
        }
    }

    public class SHAP : BaseData
    {
        public SHAP(byte[] data)
        {
            InitParams(data);
        }
    }

    public class TSTA : BaseData
    {
        public TSTAData[]               Data                    { get; set; }

        public TSTA(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<TSTAData> GenerateData
        {
            get
            {
                var data= new TSTAData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0960:
                    case 0x0961:
                    case 0x0962:
                    case 0x4A63:
                    case 0x0964:
                    case 0x0965:
                    case 0x0966:
                    case 0x0967:
                    case 0x0968:
                    case 0x0969:
                    case 0x0A6A:
                    case 0x0A6B:
                    case 0x026C: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new TSTAData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class TSTAData : ParamSet
        {
            public uint                 _0960                   { get { return GetParamAs<uint>    (0x0960); } set { SetParam(0x0960, value); } }
            public uint                 _0961                   { get { return GetParamAs<uint>    (0x0961); } set { SetParam(0x0961, value); } }
            public uint                 _0962                   { get { return GetParamAs<uint>    (0x0962); } set { SetParam(0x0962, value); } }
            public Vector3              _4A63                   { get { return GetParamAs<Vector3> (0x4A63); } set { SetParam(0x4A63, value); } }
            public uint                 _0964                   { get { return GetParamAs<uint>    (0x0964); } set { SetParam(0x0964, value); } }
            public uint                 _0965                   { get { return GetParamAs<uint>    (0x0965); } set { SetParam(0x0965, value); } }
            public uint                 _0966                   { get { return GetParamAs<uint>    (0x0966); } set { SetParam(0x0966, value); } }
            public uint                 _0967                   { get { return GetParamAs<uint>    (0x0967); } set { SetParam(0x0967, value); } }
            public uint                 _0968                   { get { return GetParamAs<uint>    (0x0968); } set { SetParam(0x0968, value); } }
            public uint                 _0969                   { get { return GetParamAs<uint>    (0x0969); } set { SetParam(0x0969, value); } }
            public float                _0A6A                   { get { return GetParamAs<float>   (0x0A6A); } set { SetParam(0x0A6A, value); } }
            public float                _0A6B                   { get { return GetParamAs<float>   (0x0A6B); } set { SetParam(0x0A6B, value); } }
            public string               _026C                   { get { return GetParamAs<string>  (0x026C); } set { SetParam(0x026C, value); } }
        }
    }

    public class TSET : BaseData
    {
        public TSETData[]               Data                    { get; set; }

        public TSET(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<TSETData> GenerateData
        {
            get
            {
                var data= new TSETData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0970:
                    case 0x0871:
                    case 0x0972:
                    case 0x0973:
                    case 0x0974:
                    case 0x8875: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new TSETData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class TSETData : ParamSet
        {
            public uint                 _0970                   { get { return GetParamAs<uint>    (0x0970); } set { SetParam(0x0970, value); } }
            public int                  _0871                   { get { return GetParamAs<int>     (0x0871); } set { SetParam(0x0871, value); } }
            public uint                 _0972                   { get { return GetParamAs<uint>    (0x0972); } set { SetParam(0x0972, value); } }
            public uint                 _0973                   { get { return GetParamAs<uint>    (0x0973); } set { SetParam(0x0973, value); } }
            public uint                 _0974                   { get { return GetParamAs<uint>    (0x0974); } set { SetParam(0x0974, value); } }
            public uint[]               _8875                   { get { return GetParamAs<uint[]>  (0x8875); } set { SetParam(0x8875, value); } }
        }
    }

    public class TEXF : BaseData
    {
        public TEXFData[]               Data                    { get; set; }

        public TEXF(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<TEXFData> GenerateData
        {
            get
            {
                var data= new TEXFData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0280: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new TEXFData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class TEXFData : ParamSet
        {
            public string               _0280                   { get { return GetParamAs<string>  (0x0280); } set { SetParam(0x0280, value); } }
        }
    }

    public class UNRM : BaseData
    {
        public UNRMData[]               Data                    { get; set; }

        public UNRM(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<UNRMData> GenerateData
        {
            get
            {
                var data= new UNRMData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x09DA:
                    case 0x89DB:
                    case 0x09DC:
                    case 0x89DD:
                    case 0x89DE: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new UNRMData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class UNRMData : ParamSet
        {
            public uint                 _09DA                   { get { return GetParamAs<uint>    (0x09DA); } set { SetParam(0x09DA, value); } }
            public uint[]               _89DB                   { get { return GetParamAs<uint[]>  (0x89DB); } set { SetParam(0x89DB, value); } }
            public uint                 _09DC                   { get { return GetParamAs<uint>    (0x09DC); } set { SetParam(0x09DC, value); } }
            public uint[]               _89DD                   { get { return GetParamAs<uint[]>  (0x89DD); } set { SetParam(0x89DD, value); } }
            public uint[]               _89DE                   { get { return GetParamAs<uint[]>  (0x89DE); } set { SetParam(0x89DE, value); } }
        }
    }

    public class NDTR : BaseData
    {
        public int                      _0801                   { get { return GetParamAs<int>     (0x0801); } set { SetParam(0x0801, value); } }
        public int                      _0802                   { get { return GetParamAs<int>     (0x0802); } set { SetParam(0x0802, value); } }
        public int                      _08FA                   { get { return GetParamAs<int>     (0x08FA); } set { SetParam(0x08FA, value); } }

        public NDTR(byte[] data)
        {
            InitParams(data);
        }
    }

    public class NODE : BaseData
    {
        public NODEData[]               Data                    { get; set; }

        public NODE(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<NODEData> GenerateData
        {
            get
            {
                var data= new NODEData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x0903:
                    case 0x0804:
                    case 0x080F:
                    case 0x0805:
                    case 0x0806:
                    case 0x4A07:
                    case 0x4A08:
                    case 0x4A09:
                    case 0xCA0A:
                    case 0x090B:
                    case 0x090C:
                    case 0x020D: data.Params.Add(i);                        break;
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new NODEData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class NODEData : ParamSet
        {
            public uint                 _0903                   { get { return GetParamAs<uint>    (0x0903); } set { SetParam(0x0903, value); } }
            public int                  _0804                   { get { return GetParamAs<int>     (0x0804); } set { SetParam(0x0804, value); } }
            public int                  _080F                   { get { return GetParamAs<int>     (0x080F); } set { SetParam(0x080F, value); } }
            public int                  _0805                   { get { return GetParamAs<int>     (0x0805); } set { SetParam(0x0805, value); } }
            public int                  _0806                   { get { return GetParamAs<int>     (0x0806); } set { SetParam(0x0806, value); } }
            public Vector3              _4A07                   { get { return GetParamAs<Vector3> (0x4A07); } set { SetParam(0x4A07, value); } }
            public Vector3              _4A08                   { get { return GetParamAs<Vector3> (0x4A08); } set { SetParam(0x4A08, value); } }
            public Vector3              _4A09                   { get { return GetParamAs<Vector3> (0x4A09); } set { SetParam(0x4A09, value); } }
            public Matrix               _CA0A                   { get { return GetParamAs<Matrix>  (0xCA0A); } set { SetParam(0xCA0A, value); } }
            public uint                 _090B                   { get { return GetParamAs<uint>    (0x090B); } set { SetParam(0x090B, value); } }
            public uint                 _090C                   { get { return GetParamAs<uint>    (0x090C); } set { SetParam(0x090C, value); } }
            public string               _020D                   { get { return GetParamAs<string>  (0x020D); } set { SetParam(0x020D, value); } }
        }
    }

    public class NODO : BaseData
    {
        public NODOData[]               Data                    { get; set; }

        public NODO(byte[] data)
        {
            InitParams(data);

            Data    = GenerateData.ToArray();
        }

        public IEnumerable<NODOData> GenerateData
        {
            get
            {
                var data= new NODOData();

                foreach(var i in Params)
                {
                    switch(i.ID)
                    {
                    case 0x00FC:                                            break;
                    case 0x00FD:                                            break;
                  //case 0x00FD: yield return data;                         break;
                    case 0x00FE: yield return data; data= new NODOData();   break;
                    default: throw new Exception();
                    }
                }

                yield return data;
            }
        }

        public class NODOData : ParamSet
        {
        }
    }
}
