﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace Models.Aqo
{
    public class AFPReader : IDisposable
    {
        public BinaryReader             Reader                  { get; set; }

        public AFPReader(string file)
        {
            Reader = new BinaryReader(File.OpenRead(file));
        }

        public AFPReader(Stream s)
        {
            Reader = new BinaryReader(s);
        }

        public void Dispose()
        {
            if(null != Reader)
            {
                Reader.Dispose();

                Reader  = null;
            }
        }

        public AFPFile Read()
        {
            var r   = Reader;
            var data= new AFPFile();
            var type= LoadHeaders(data);

            if(Reader.BaseStream is FileStream)
                data.FileName   = ((FileStream)Reader.BaseStream).Name;

            switch(type)
            {
            case "VTBF":
                data.VTBF       = type;
                data.Unknown1   = r.ReadUInt32();
                data.AQGF       = Encoding.ASCII.GetString(r.ReadBytes(4));
                data.Unknown2   = r.ReadUInt32();

                Assert(data.Unknown2 == 0x4C000001);

                LoadChunks(data);
                //System.Diagnostics.Debug.Print("VTBF: {0}", file);
                break;

            default:
                throw new FormatException("無効なファイル形式です");
            }

            data.Dictionary = data.AsDictionary();

            return data;
        }

        private string LoadHeaders(AFPFile data)
        {
            var r   = Reader;
            var type= Encoding.ASCII.GetString(r.ReadBytes(4));

            if(type == "afp\0")
            {
                data.afp        = type;
                data.Unknown0   = r.ReadBytes(44);
                data.Size1      = r.ReadUInt32();
                data.Size2      = r.ReadUInt32();
                data.Size3      = r.ReadUInt32();
                type            = Encoding.ASCII.GetString(r.ReadBytes(4));
            }

            if(type == "aqo\0")
            {
                data.aqo        = type;
                type            = Encoding.ASCII.GetString(r.ReadBytes(4));
            }

            return type;
        }

        private void LoadChunks(AFPFile data)
        {
            var r   = Reader;
            var s   = r.BaseStream;
            var size= data.Size1 == 0
                    ? r.BaseStream.Length
                    : data.Size1 + r.BaseStream.Position - 0x10;

            while(s.Position < size - 16)
            {
                var vtc0    = new vtc0();
                vtc0.Position= s.Position;
                vtc0.Type   = Encoding.ASCII.GetString(r.ReadBytes(4));
                vtc0.Size   = r.ReadInt32();
                vtc0.SubType= Encoding.ASCII.GetString(r.ReadBytes(4));

                Assert(vtc0.Type　== "vtc0");
                Assert(vtc0.Size >= 4 && vtc0.Size < 0x10000000);

                vtc0.Data   = r.ReadBytes(vtc0.Size - 4);

                //System.Diagnostics.Debug.Print("  TYPE:{0}", vtc0.SubType);

                switch(vtc0.SubType)
                {
                case "ROOT":    vtc0.DataObj= new ROOT(vtc0.Data);  break;
                case "OBJC":    vtc0.DataObj= new OBJC(vtc0.Data);  break;
                case "VSET":    vtc0.DataObj= new VSET(vtc0.Data);  break;
                case "VTXE":    vtc0.DataObj= new VTXE(vtc0.Data);  break;
                case "VTXL":    vtc0.DataObj= new VTXL(vtc0.Data);  break;
                case "PSET":    vtc0.DataObj= new PSET(vtc0.Data);  break;
                case "MESH":    vtc0.DataObj= new MESH(vtc0.Data);  break;
                case "MATE":    vtc0.DataObj= new MATE(vtc0.Data);  break;
                case "REND":    vtc0.DataObj= new REND(vtc0.Data);  break;
                case "SHAD":    vtc0.DataObj= new SHAD(vtc0.Data);  break;
                case "SHAP":    vtc0.DataObj= new SHAP(vtc0.Data);  break;
                case "TSTA":    vtc0.DataObj= new TSTA(vtc0.Data);  break;
                case "TSET":    vtc0.DataObj= new TSET(vtc0.Data);  break;
                case "TEXF":    vtc0.DataObj= new TEXF(vtc0.Data);  break;
                case "UNRM":    vtc0.DataObj= new UNRM(vtc0.Data);  break;
                case "NDTR":    vtc0.DataObj= new NDTR(vtc0.Data);  break;
                case "NODE":    vtc0.DataObj= new NODE(vtc0.Data);  break;
                case "NODO":    vtc0.DataObj= new NODO(vtc0.Data);  break;
                default:    throw new Exception();
                }

                data.VTC0.Add(vtc0);
            }
        }

        public static void Assert(bool b)
        {
            if(!b)
                throw new Exception();
        }
    }
}
