﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SlimDX;

namespace Models.Aqo
{
    using Vertex = Tuple<Vector3, Vector3, Vector3, Vector3, uint, Vector4>;

    public class ImportMesh
    {
        public Dictionary<int, int>     Bones                   { get; set; }
        public int                      IndexMATE               { get; set; }
        public int                      IndexREND               { get; set; }
        public int                      IndexSHAD               { get; set; }
        public int                      IndexTSET               { get; set; }
        public MeshBuilder<Vertex>      MeshBuilder             { get; set; }
        public bool                     Skinning                { get; set; }

        public ImportMesh()
        {
            MeshBuilder = new MeshBuilder<Vertex>();
            Bones       = new Dictionary<int, int>();
            Skinning    = false;
        }

        public bool CanAdd(IEnumerable<ushort> bones)
        {
            int n   = 0;

            foreach(var i in bones)
            {
                if(Bones.ContainsKey(i))
                    continue;

                ++n;
            }

            if(Bones.Count + n <= 16)
                return true;

            return false;
        }

        public void AddFace(
            Vector3 p0, Vector3 n0, Vector3 t0, ImportSkin s0,
            Vector3 p1, Vector3 n1, Vector3 t1, ImportSkin s1,
            Vector3 p2, Vector3 n2, Vector3 t2, ImportSkin s2)
        {
            AddVertex(p0, n0, t0, s0);
            AddVertex(p1, n1, t1, s1);
            AddVertex(p2, n2, t2, s2);
        }

        public void AddVertex(Vector3 p, Vector3 n, Vector3 t, ImportSkin s)
        {
            AddBones(s.Bones);

          //MeshBuilder.Add(Tuple.Create(p, n, t, t, BitConverter.ToUInt32(GetBoneIndices(s), 0), GetBoneWeights(s)));
            MeshBuilder.Add(Tuple.Create(p, n, t, s.UV1, BitConverter.ToUInt32(GetBoneIndices(s), 0), GetBoneWeights(s)));
        }

        public void AddBones(IEnumerable<ushort> bones)
        {
            foreach(var i in bones)
                AddBone(i);
        }

        public void AddBone(ushort bone)
        {
            if(!Bones.ContainsKey(bone))
            {
                System.Diagnostics.Debug.Assert(Bones.Count < 16);

                Bones.Add(bone, Bones.Count);
            }
        }

        public byte[] GetBoneIndices(ImportSkin s)
        {
            return new byte[]
            {
                (byte)(s.W0 == 0 ? 0 : Bones[s.B0]),
                (byte)(s.W1 == 0 ? 0 : Bones[s.B1]),
                (byte)(s.W2 == 0 ? 0 : Bones[s.B2]),
                (byte)(s.W3 == 0 ? 0 : Bones[s.B3]),
            };
        }

        public Vector4 GetBoneWeights(ImportSkin s)
        {
            return new Vector4(s.W0, s.W1, s.W2, s.W3);
        }
    }

    public class ImportMeshSet
    {
        public List<ImportMesh>         Meshes                  { get; set; }

        public ImportMeshSet()
        {
            Meshes  = new List<ImportMesh>();
        }

        public void AddFace(
            ImportSkinSet skins,
            Vector3 p0, Vector3 n0, Vector3 t0,
            Vector3 p1, Vector3 n1, Vector3 t1,
            Vector3 p2, Vector3 n2, Vector3 t2)
        {
            var mesh    = (ImportMesh)null;
            var s0      = skins.Get(p0);
            var s1      = skins.Get(p1);
            var s2      = skins.Get(p2);
            var bones   = s0.Bones.Union(s1.Bones).Union(s2.Bones).Distinct().ToArray();

            foreach(var i in Meshes)
            {
                if(i.CanAdd(bones))
                {
                    mesh= i;
                    break;
                }
            }

            if(null == mesh)
                Meshes.Add(mesh= new ImportMesh());

            mesh.AddFace(
                p0, n0, t0, s0,
                p1, n1, t1, s1,
                p2, n2, t2, s2);
        }
    }

    public class ImportSkinSet
    {
        public Dictionary<Vector3, ImportSkin> Skins            { get; set; }

        public ImportSkinSet()
        {
            Skins   = new Dictionary<Vector3, ImportSkin>();
        }

        public void Create(VertexAccessor va, int numvert, ushort[] bones)
        {
            for(int j= 0; j < numvert; ++j)
            {
                var pos = new Vector3(va.PX[j], va.PY[j], va.PZ[j]);

                if(Skins.ContainsKey(pos))
                    continue;

                // ボーンの番号はモデル全体での番号に変換する
                if(null == va.U1)
                {
                    Add(pos,
                        bones[va.B0[j]], va.W0[j],
                        bones[va.B1[j]], va.W1[j],
                        bones[va.B2[j]], va.W2[j],
                        bones[va.B3[j]], va.W3[j]);
                } else
                {
                    Add(pos,
                        bones[va.B0[j]], va.W0[j],
                        bones[va.B1[j]], va.W1[j],
                        bones[va.B2[j]], va.W2[j],
                        bones[va.B3[j]], va.W3[j],
                        va.U1[j], va.V1[j]);
                }
            }
        }

        public void Add(Vector3 pos, ImportSkin skin)
        {
            Skins[pos]  = skin;
        }

        public void Add(Vector3 pos,
            ushort b0, float w0,
            ushort b1, float w1,
            ushort b2, float w2,
            ushort b3, float w3,
            float u1 = 0, float v1 = 0)
        {
            Skins[pos]  = new ImportSkin()
            {
                Position= pos,
                B0      = b0,
                B1      = b1,
                B2      = b2,
                B3      = b3,
                W0      = w0,
                W1      = w1,
                W2      = w2,
                W3      = w3,
                UV1     = new Vector3(u1, v1, 0),
            };
        }

        public ImportSkin Get(Vector3 pos)
        {
            var maxd= float.MaxValue;
            var skin= (ImportSkin)null;
          //var list= new List<Tuple<float, ImportSkin>>();

            foreach(var i in Skins.Values)
            {
                var d   = (i.Position - pos).LengthSquared();

                if(d < maxd)
                {
                    maxd= d;
                    skin= i;
                    /*
                    list.Add(Tuple.Create(d, i));

                    if(list.Count > 4)
                        list.RemoveAt(0);
                    */
                }
            }

            #if false
            list.Reverse();

            if(list.First().Item1 < 0.001f)
                    return list.First().Item2;
            else    return new VertexSkin(p, list).CreateSkin();

            //System.Diagnostics.Debug.Print("Closest: {0} => {1:F8}", p, maxd);
            #else
            return skin;
            #endif
        }
    }

    public class ImportSkin
    {
        public Vector3                  Position;
        public ushort                   B0, B1, B2, B3;
        public float                    W0, W1, W2, W3;
        public Vector3                  UV1;
        public IEnumerable<ushort>      Bones
        {
            get
            {
                if(W0 > 0)  yield return B0;
                if(W1 > 0)  yield return B1;
                if(W2 > 0)  yield return B2;
                if(W3 > 0)  yield return B3;
            }
        }

        public override string ToString()
        {
            return string.Format("{0}:{1:F8} {2}:{3:F8} {4}:{5:F8} {6}:{7:F8}",
                B0, W0, B1, W1, B2, W2, B3, W3);
        }
    }
}
